from witnet_lib.witnet_client import WitnetClient
from witnet_lib.logger import log
from queue import Queue
import threading




class MapNodes():
    def __init__(self, config, peer_addrs):
        self.peer_addrs = peer_addrs
        self.config = config

        # https://www.geeksforgeeks.org/queue-in-python/
        self.queue = Queue(maxsize=0)
        for peer_addr in peer_addrs: 
            self.queue.put(peer_addr)
        self.visited_nodes = set()
        self.active = {}
        # https://www.educative.io/edpresso/what-are-locks-in-python
        self.active_set_lock = threading.Lock()
        self.queue_lock = threading.Lock()
        self.set_lock = threading.Lock()

    def worker(self, worker_num):
        log.info(f"Starting worker {worker_num}")

        while True:
            try:
                # https://stackoverflow.com/questions/21320621/python-queue-block-timeout-does-not-timeout-any-idea-why
                peer = self.queue.get(block=True, timeout=5)
            except:
                return

            log.info(f"worker {worker_num}: peer {peer}")
            if peer in self.visited_nodes:
                continue
            # create client
            client = WitnetClient(self.config)
            try:
                client.handshake(peer)
                peers = client.get_peers()
                client.close()
                self.active_set_lock.acquire()
                self.active[peer]=peers
                self.active_set_lock.release()
            except:
                peers = []
            log.info(peers)

            # queue lock
            self.queue_lock.acquire()
            print(f"Acquiring queue lock: worker {worker_num}")
            for p in peers:
                if p not in self.visited_nodes:
                    self.queue.put(p)
            self.queue_lock.release()

            # set lock
            self.set_lock.acquire()
            print(f"Acquiring set lock: worker {worker_num}")
            self.visited_nodes.add(peer)
            self.set_lock.release()

    def start_mapping_workers(self, num):
        threads = []
        for i in  range(num):
            threads.append ( threading.Thread(target=self.worker, args=(i, ) ) )
        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()
        return self.visited_nodes, self.active


from datetime import datetime
import random
from witnet_lib.proto_lib import witnet_pb2
from witnet_lib.logger import log

def byte_ip(addr):
    ip, port = addr.split(":")
    port = int(port)
    ip = [int(i) for i in ip.split(".")]
    # print(hex(port))
    port =  list(bytes.fromhex(hex(port)[2:]))
    l = ip+port
    return bytes(l)

class WitnetMsgHandler():
    def __init__(self, config):
        self.magic = config.magic
        self.genesis_sec = config.genesis_sec
        self.sender_addr = config.sender_addr
        self.time_per_epoch = config.time_per_epoch
    def version_cmd(self, receiver_addr):
        now = datetime.now()
        utc_sec = now.strftime('%s')
        utc_sec = int(utc_sec)
        
        rece_addr_msg = witnet_pb2.Address()

        rece_addr_msg.address = byte_ip(self.sender_addr)
        send_addr_msg = witnet_pb2.Address()
        send_addr_msg.address = byte_ip(receiver_addr)
        version = witnet_pb2.Version()
        version.sender_address.CopyFrom(send_addr_msg)
        version.receiver_address.CopyFrom(rece_addr_msg)
        version.version = 1 
        version.capabilities = 1
        version.user_agent = b"full-node-desktop-edition".decode('utf-8')
        version.timestamp = utc_sec
        version.last_epoch = int((utc_sec - self.genesis_sec)/self.time_per_epoch) -1
        version.nonce = random.randint(10000000,100000000000000000) 
        
        cmd = witnet_pb2.Message.Command()
        cmd.Version.CopyFrom(version)
        return cmd
    
    def verack_cmd(self):
        verack = witnet_pb2.Verack()
        cmd = witnet_pb2.Message.Command()
        cmd.Verack.CopyFrom(verack)
        return cmd

    def get_peers_cmd(self):
        get_peers = witnet_pb2.GetPeers()
        cmd = witnet_pb2.Message.Command()
        cmd.GetPeers.CopyFrom(get_peers)
        return cmd

    def serialize(self, cmd):
        msg = witnet_pb2.Message()
        msg.magic = self.magic
        msg.kind.CopyFrom(cmd)
        serialized = msg.SerializeToString()
        bytes_len = len(serialized)
        length = bytes_len.to_bytes(4, byteorder='big')
        msg = length + serialized
        return msg

    def parse_msg(self, rece_bytes):
        msg = witnet_pb2.Message()
        msg.ParseFromString(rece_bytes)
        log.debug(msg)
        return msg
    
    def parse_peers(self, msg):
        peers = []
        if msg.kind.HasField("Peers"):
            for peer in msg.kind.Peers.peers:
                addr = peer.address
                ip, port = addr[:4], addr[4:]
                # from bytes convert to ip and port
                # data transferred over network in big endian
                ip = '.'.join([str(byt) for byt in ip])
                port = int.from_bytes(port, byteorder='big')

                peers.append(f"{ip}:{port}")
        
        return peers
        
        



from witnet_lib.logger import log
from witnet_lib.proto_lib.witnet_msg import WitnetMsgHandler
from witnet_lib.tcp_handler import TCPSocket

class WitnetClient():
    
    def __init__(self,config):
        self.config = config
        self.msg_handler = WitnetMsgHandler(self.config)

    def handshake(self, peer_addr):
        # connect to peer 
        self.tcp_handler = TCPSocket("connect")
        self.tcp_handler.connect(peer_addr)

        # send version to node
        version_cmd = self.msg_handler.version_cmd(peer_addr)
        version_msg = self.msg_handler.serialize(version_cmd)
        self.tcp_handler.send(version_msg)
        
        # receive verack from node
        self.tcp_handler.receive_witnet_msg()
        # receive version from node
        self.tcp_handler.receive_witnet_msg()

        # send verack to node
        verack_cmd = self.msg_handler.verack_cmd()
        verack_msg = self.msg_handler.serialize(verack_cmd)
        self.tcp_handler.send(verack_msg)

    def get_peers(self):
        # send get peer request to node
        get_peers_cmd = self.msg_handler.get_peers_cmd()
        get_peers_msg = self.msg_handler.serialize(get_peers_cmd)
        self.tcp_handler.send(get_peers_msg)
        i = 0
        while i<10:
            msg = self.tcp_handler.receive_witnet_msg()
            parsed_msg = self.msg_handler.parse_msg(msg)
            peers = self.msg_handler.parse_peers(parsed_msg)
            if len(peers) > 0 :
                return peers
            i+=1
        return []
    def close(self):
        self.tcp_handler.close()



import socket
from datetime import datetime
from witnet_lib.logger import log
from witnet_lib.utils import resolve_url
import random
import threading
import time


class TCPSocket:
    sock_type = "connect"

    def __init__(self, sock_type="connect"):
        self.sock_type = sock_type

    def connect(self, node_addr):
        self.node_addr = node_addr
        host, port = resolve_url(node_addr)
        if self.sock_type == "connect":
            log.info(f"Connecting to {host}:{port}")
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((host, port))

    def send(self, msg):
        log.debug(f"Sending [>] {msg}")
        try:
            self.sock.send(msg)
        except:
            # recreate the socket and reconnect
            self.connect(self.node_addr)
            self.sock.send(msg)

    def receive(self, length=4096):
        if self.sock.fileno() == -1:
            self.connect(self.node_addr)
        msg = self.sock.recv(length)
        log.debug(f"Receiving [<] {msg}")
        return msg

    def receive_witnet_msg(self):
        length = self.receive(4)
        rece_len = int.from_bytes(length, byteorder='big')
        return self.receive(rece_len)

    def close(self):
        self.sock.close()

    def listen(self, port):
        if self.sock_type == "listen":
            self.sock = socket.socket()
            self.sock.bind(('', port))
            self.sock.listen(5)
            log.info(f"Listening at {port}")
